--
-- This file is part of TALER
-- Copyright (C) 2025 Taler Systems SA
--
-- TALER is free software; you can redistribute it and/or modify it under the
-- terms of the GNU General Public License as published by the Free Software
-- Foundation; either version 3, or (at your option) any later version.
--
-- TALER is distributed in the hope that it will be useful, but WITHOUT ANY
-- WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
-- A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License along with
-- TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>

SELECT _v.register_patch('magnet-bank-0001', NULL, NULL);

CREATE SCHEMA magnet_bank;
SET search_path TO magnet_bank;

CREATE TYPE taler_amount AS (val INT8, frac INT4);
COMMENT ON TYPE taler_amount IS 'Stores an amount, fraction is in units of 1/100000000 of the base value';

CREATE TABLE tx_in(
  tx_in_id INT8 PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
  magnet_code INT8 UNIQUE,
  amount taler_amount NOT NULL,
  subject TEXT NOT NULL,
  debit_account TEXT NOT NULL,
  debit_name TEXT NOT NULL,
  valued_at INT8 NOT NULL,
  registered_at INT8 NOT NULL
);
COMMENT ON TABLE tx_in IS 'Incoming transactions';

CREATE TABLE tx_out(
  tx_out_id INT8 PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
  magnet_code INT8 UNIQUE,
  amount taler_amount NOT NULL,
  subject TEXT NOT NULL,
  credit_account TEXT NOT NULL,
  credit_name TEXT NOT NULL,
  valued_at INT8 NOT NULL,
  registered_at INT8 NOT NULL
);
COMMENT ON TABLE tx_in IS 'Outgoing transactions';

CREATE TYPE incoming_type AS ENUM
  ('reserve' ,'kyc', 'wad');
COMMENT ON TYPE incoming_type IS 'Types of incoming talerable transactions';

CREATE TABLE taler_in(
  tx_in_id INT8 PRIMARY KEY REFERENCES tx_in(tx_in_id) ON DELETE CASCADE,
  type incoming_type NOT NULL,
  metadata BYTEA NOT NULL,
  origin_exchange_url TEXT,
  CONSTRAINT polymorphism CHECK(
    CASE type
      WHEN 'wad' THEN LENGTH(metadata)=24 AND origin_exchange_url IS NOT NULL
      ELSE LENGTH(metadata)=32 AND origin_exchange_url IS NULL
    END
  )
);
COMMENT ON TABLE tx_in IS 'Incoming talerable transactions';

CREATE UNIQUE INDEX taler_in_unique_reserve_pub ON taler_in (metadata) WHERE type = 'reserve';

CREATE TABLE taler_out(
  tx_out_id INT8 PRIMARY KEY REFERENCES tx_out(tx_out_id) ON DELETE CASCADE,
  wtid BYTEA NOT NULL UNIQUE CHECK (LENGTH(wtid)=32),
  exchange_base_url TEXT NOT NULL
);
COMMENT ON TABLE tx_in IS 'Outgoing talerable transactions';

CREATE TYPE transfer_status AS ENUM(
  'pending',
  'transient_failure',
  'permanent_failure',
  'success',
  'late_failure'
);
COMMENT ON TYPE transfer_status IS 'Status of an initiated outgoing transaction';

CREATE TABLE initiated(
  initiated_id INT8 PRIMARY KEY GENERATED ALWAYS AS IDENTITY,
  amount taler_amount NOT NULL,
  subject TEXT NOT NULL,
  credit_account TEXT NOT NULL,
  credit_name TEXT NOT NULL,
  status transfer_status NOT NULL DEFAULT 'pending',
  status_msg TEXT,
  magnet_code INT8 UNIQUE,
  last_submitted INT8,
  submission_counter INT2 NOT NULL DEFAULT 0,
  tx_out_id INT8 UNIQUE REFERENCES tx_out(tx_out_id) ON DELETE CASCADE,
  initiated_at INT8 NOT NULL
);
COMMENT ON TABLE tx_in IS 'Initiated outgoing transactions';

CREATE TABLE transfer(
  initiated_id INT8 PRIMARY KEY REFERENCES initiated(initiated_id) ON DELETE CASCADE,
  request_uid BYTEA UNIQUE NOT NULL CHECK (LENGTH(request_uid)=64),
  wtid BYTEA UNIQUE NOT NULL CHECK (LENGTH(wtid)=32),
  exchange_base_url TEXT NOT NULL
);
COMMENT ON TABLE transfer IS 'Wire Gateway transfers';

CREATE TABLE bounced(
  tx_in_id INT8 NOT NULL UNIQUE REFERENCES tx_in(tx_in_id) ON DELETE CASCADE,
  initiated_id INT8 NOT NULL UNIQUE REFERENCES initiated(initiated_id) ON DELETE CASCADE,
  reason TEXT NOT NULL
);
COMMENT ON TABLE tx_in IS 'Bounced transactions';