--
-- This file is part of TALER
-- Copyright (C) 2025 Taler Systems SA
--
-- TALER is free software; you can redistribute it and/or modify it under the
-- terms of the GNU General Public License as published by the Free Software
-- Foundation; either version 3, or (at your option) any later version.
--
-- TALER is distributed in the hope that it will be useful, but WITHOUT ANY
-- WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
-- A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License along with
-- TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>

SET search_path TO magnet_bank;

-- Remove all existing functions
DO
$do$
DECLARE
  _sql text;
BEGIN
  SELECT INTO _sql
        string_agg(format('DROP %s %s CASCADE;'
                        , CASE prokind
                            WHEN 'f' THEN 'FUNCTION'
                            WHEN 'p' THEN 'PROCEDURE'
                          END
                        , oid::regprocedure)
                  , E'\n')
  FROM   pg_proc
  WHERE  pronamespace = 'magnet_bank'::regnamespace;

  IF _sql IS NOT NULL THEN
    EXECUTE _sql;
  END IF;
END
$do$;

CREATE FUNCTION register_tx_in(
  IN in_code INT8,
  IN in_amount taler_amount,
  IN in_subject TEXT,
  IN in_debit_account TEXT,
  IN in_debit_name TEXT,
  IN in_valued_at INT8,
  IN in_type incoming_type,
  IN in_metadata BYTEA,
  IN in_now INT8,
  -- Error status
  OUT out_reserve_pub_reuse BOOLEAN,
  -- Success return
  OUT out_tx_row_id INT8,
  OUT out_valued_at INT8,
  OUT out_new BOOLEAN
)
LANGUAGE plpgsql AS $$
BEGIN
-- Check for idempotence
SELECT tx_in_id, valued_at
INTO out_tx_row_id, out_valued_at
FROM tx_in
WHERE (in_code IS NOT NULL AND magnet_code = in_code) -- Magnet transaction
  OR (in_code IS NULL AND amount = in_amount AND debit_account = in_debit_account AND subject = in_subject); -- Admin transaction
out_new = NOT found;
IF NOT out_new THEN
  out_reserve_pub_reuse=false;
  RETURN;
END IF;

-- Check conflict
SELECT in_type = 'reserve'::incoming_type AND EXISTS(SELECT FROM taler_in WHERE metadata = in_metadata AND type = 'reserve')
  INTO out_reserve_pub_reuse;
IF out_reserve_pub_reuse THEN
  RETURN;
END IF;

-- Insert new incoming transaction
out_valued_at = in_valued_at;
INSERT INTO tx_in (
  magnet_code,
  amount,
  subject,
  debit_account,
  debit_name,
  valued_at,
  registered_at
) VALUES (
  in_code,
  in_amount,
  in_subject,
  in_debit_account,
  in_debit_name,
  in_valued_at,
  in_now
)
RETURNING tx_in_id INTO out_tx_row_id;
-- Notify new incoming transaction registration
PERFORM pg_notify('tx_in', out_tx_row_id || '');
IF in_type IS NOT NULL THEN
  -- Insert new incoming talerable transaction
  INSERT INTO taler_in (
    tx_in_id,
    type,
    metadata
  ) VALUES (
    out_tx_row_id,
    in_type,
    in_metadata
  );
  -- Notify new incoming talerable transaction registration
  PERFORM pg_notify('taler_in', out_tx_row_id || '');
END IF;
END $$;
COMMENT ON FUNCTION register_tx_in IS 'Register an incoming transaction idempotently';

CREATE FUNCTION register_tx_out(
  IN in_code INT8,
  IN in_amount taler_amount,
  IN in_subject TEXT,
  IN in_credit_account TEXT,
  IN in_credit_name TEXT,
  IN in_valued_at INT8,
  IN in_wtid BYTEA,
  IN in_origin_exchange_url TEXT,
  IN in_now INT8,
  -- Success return
  OUT out_tx_row_id INT8,
  OUT out_new BOOLEAN
)
LANGUAGE plpgsql AS $$
BEGIN
-- Check for idempotence
SELECT tx_out_id INTO out_tx_row_id
FROM tx_out WHERE magnet_code = in_code;

out_new = NOT found;
IF out_new THEN
  -- Insert new outgoing transaction
  INSERT INTO tx_out (
    magnet_code,
    amount,
    subject,
    credit_account,
    credit_name,
    valued_at,
    registered_at
  ) VALUES (
    in_code,
    in_amount,
    in_subject,
    in_credit_account,
    in_credit_name,
    in_valued_at,
    in_now
  )
  RETURNING tx_out_id INTO out_tx_row_id;
  -- Notify new outgoing transaction registration
  PERFORM pg_notify('tx_out', out_tx_row_id || '');

  IF in_wtid IS NOT NULL THEN
    -- Insert new outgoing talerable transaction
    INSERT INTO taler_out (
      tx_out_id,
      wtid,
      exchange_base_url
    ) VALUES (
      out_tx_row_id,
      in_wtid,
      in_origin_exchange_url
    );
    -- Notify new outgoing talerable transaction registration
    PERFORM pg_notify('taler_out', out_tx_row_id || '');
  END IF;
END IF;
END $$;
COMMENT ON FUNCTION register_tx_out IS 'Register an outgoing transaction idempotently';

CREATE FUNCTION taler_transfer(
  IN in_request_uid BYTEA,
  IN in_wtid BYTEA,
  IN in_subject TEXT,
  IN in_amount taler_amount,
  IN in_exchange_base_url TEXT,
  IN in_credit_account TEXT,
  IN in_credit_name TEXT,
  IN in_now INT8,
  -- Error return
  OUT out_request_uid_reuse BOOLEAN,
  OUT out_wtid_reuse BOOLEAN,
  -- Success return
  OUT out_initiated_row_id INT8,
  OUT out_initiated_at INT8
)
LANGUAGE plpgsql AS $$
BEGIN
-- Check for idempotence and conflict
SELECT (amount != in_amount 
          OR credit_account != in_credit_account
          OR exchange_base_url != in_exchange_base_url
          OR wtid != in_wtid)
        ,initiated_id, initiated_at
INTO out_request_uid_reuse, out_initiated_row_id, out_initiated_at
FROM transfer JOIN initiated USING (initiated_id)
WHERE request_uid = in_request_uid;
IF FOUND THEN
  out_wtid_reuse=FALSE;
  RETURN;
END IF;
out_request_uid_reuse=FALSE;
-- Check for wtid reuse
out_wtid_reuse = EXISTS(SELECT FROM transfer WHERE wtid=in_wtid);
IF out_wtid_reuse THEN
  RETURN;
END IF;
-- Insert an initiated outgoing transaction
out_initiated_at = in_now;
INSERT INTO initiated (
  amount,
  subject,
  credit_account,
  credit_name,
  initiated_at
) VALUES (
  in_amount,
  in_subject,
  in_credit_account,
  in_credit_name,
  in_now
) RETURNING initiated_id 
INTO out_initiated_row_id;
-- Insert a transfer operation
INSERT INTO transfer (
  initiated_id,
  request_uid,
  wtid,
  exchange_base_url
) VALUES (
  out_initiated_row_id,
  in_request_uid,
  in_wtid,
  in_exchange_base_url
);
PERFORM pg_notify('transfer', out_initiated_row_id || '');
END $$;

CREATE FUNCTION initiated_status_update(
  IN in_initiated_id INT8,
  IN in_status transfer_status,
  IN in_status_msg TEXT
)
RETURNS void
LANGUAGE plpgsql AS $$
DECLARE
current_status transfer_status;
BEGIN
  -- Check current status
  SELECT status INTO current_status FROM initiated
    WHERE initiated_id = in_initiated_id;
  IF FOUND THEN
    -- Update unsettled transaction status
    IF current_status = 'success' AND in_status = 'permanent_failure' THEN
      UPDATE initiated 
      SET status = 'late_failure', status_msg = in_status_msg
      WHERE initiated_id = in_initiated_id;
    ELSIF current_status NOT IN ('success', 'permanent_failure', 'late_failure') THEN
      UPDATE initiated 
      SET status = in_status, status_msg = in_status_msg
      WHERE initiated_id = in_initiated_id;
    END IF;
  END IF;
END $$;

CREATE FUNCTION register_bounce_tx_in(
  IN in_code INT8,
  IN in_amount taler_amount,
  IN in_subject TEXT,
  IN in_debit_account TEXT,
  IN in_debit_name TEXT,
  IN in_valued_at INT8,
  IN in_bounce_amount taler_amount,
  IN in_reason TEXT,
  IN in_now INT8,
  -- Success return
  OUT out_tx_row_id INT8,
  OUT out_tx_new BOOLEAN,
  OUT out_bounce_row_id INT8,
  OUT out_bounce_new BOOLEAN
)
LANGUAGE plpgsql AS $$
BEGIN
-- Register incoming transaction idempotently
SELECT register_tx_in.out_tx_row_id, register_tx_in.out_new
INTO out_tx_row_id, out_tx_new
FROM register_tx_in(in_code, in_amount, in_subject, in_debit_account, in_debit_name, in_valued_at, NULL, NULL, in_now);

-- Check if already bounce
SELECT initiated_id
  INTO out_bounce_row_id
  FROM bounced JOIN initiated USING (initiated_id)
  WHERE tx_in_id = out_tx_row_id;
out_bounce_new=NOT FOUND;
-- Else initiate the bounce transaction
IF out_bounce_new THEN
  -- Initiate the bounce transaction
  INSERT INTO initiated (
    amount,
    subject,
    credit_account,
    credit_name,
    initiated_at
  ) VALUES (
    in_amount,
    'bounce: ' || in_code,
    in_debit_account,
    in_debit_name,
    in_now
  )
  RETURNING initiated_id INTO out_bounce_row_id;
  -- Register the bounce
  INSERT INTO bounced (
    tx_in_id,
    initiated_id,
    reason
  ) VALUES (
    out_tx_row_id,
    out_bounce_row_id,
    in_reason
  );
END IF;
END $$;
COMMENT ON FUNCTION register_bounce_tx_in IS 'Register an incoming transaction and bounce it idempotently';